#include <msp/core/hash.h>
#include <msp/core/raii.h>
#include <msp/strings/lexicalcast.h>
#include "builtin.h"
#include "generate.h"

using namespace std;

namespace Msp {
namespace GL {
namespace SL {

void DeclarationCombiner::apply(Stage &stage)
{
	stage.content.visit(*this);
	NodeRemover().apply(stage, nodes_to_remove);
}

void DeclarationCombiner::visit(Block &block)
{
	if(current_block)
		return;

	TraversingVisitor::visit(block);
}

void DeclarationCombiner::visit(VariableDeclaration &var)
{
	VariableDeclaration *&ptr = variables[var.name];
	if(ptr)
	{
		ptr->type = var.type;
		if(var.init_expression)
			ptr->init_expression = var.init_expression;
		if(var.layout)
		{
			if(ptr->layout)
			{
				for(vector<Layout::Qualifier>::iterator i=var.layout->qualifiers.begin(); i!=var.layout->qualifiers.end(); ++i)
				{
					bool found = false;
					for(vector<Layout::Qualifier>::iterator j=ptr->layout->qualifiers.begin(); (!found && j!=ptr->layout->qualifiers.end()); ++j)
						if(j->name==i->name)
						{
							j->has_value = i->value;
							j->value = i->value;
							found = true;
						}

					if(!found)
						ptr->layout->qualifiers.push_back(*i);
				}
			}
			else
				ptr->layout = var.layout;
		}
		nodes_to_remove.insert(&var);
	}
	else
		ptr = &var;
}


ConstantSpecializer::ConstantSpecializer():
	values(0)
{ }

void ConstantSpecializer::apply(Stage &stage, const map<string, int> *v)
{
	values = v;
	stage.content.visit(*this);
}

void ConstantSpecializer::visit(VariableDeclaration &var)
{
	bool specializable = false;
	if(var.layout)
	{
		vector<Layout::Qualifier> &qualifiers = var.layout->qualifiers;
		for(vector<Layout::Qualifier>::iterator i=qualifiers.begin(); i!=qualifiers.end(); ++i)
			if(i->name=="constant_id")
			{
				specializable = true;
				if(values)
					qualifiers.erase(i);
				else if(i->value==-1)
					i->value = hash32(var.name)&0x7FFFFFFF;
				break;
			}

		if(qualifiers.empty())
			var.layout = 0;
	}

	if(specializable && values)
	{
		map<string, int>::const_iterator i = values->find(var.name);
		if(i!=values->end())
		{
			RefPtr<Literal> literal = new Literal;
			if(var.type=="bool")
			{
				literal->token = (i->second ? "true" : "false");
				literal->value = static_cast<bool>(i->second);
			}
			else if(var.type=="int")
			{
				literal->token = lexical_cast<string>(i->second);
				literal->value = i->second;
			}
			var.init_expression = literal;
		}
	}
}


void BlockHierarchyResolver::enter(Block &block)
{
	block.parent = current_block;
}


TypeResolver::TypeResolver():
	stage(0)
{ }

void TypeResolver::apply(Stage &s)
{
	stage = &s;
	s.types.clear();
	s.content.visit(*this);
}

TypeDeclaration *TypeResolver::resolve_type(const string &name)
{
	map<string, TypeDeclaration *>::iterator i = stage->types.find(name);
	if(i!=stage->types.end())
	{
		map<TypeDeclaration *, TypeDeclaration *>::iterator j = alias_map.find(i->second);
		return (j!=alias_map.end() ? j->second : i->second);
	}
	else
		return 0;
}

void TypeResolver::visit(Block &block)
{
	for(NodeList<Statement>::iterator i=block.body.begin(); i!=block.body.end(); ++i)
	{
		if(!block.parent)
			type_insert_point = i;
		(*i)->visit(*this);
	}
}

void TypeResolver::visit(BasicTypeDeclaration &type)
{
	type.base_type = resolve_type(type.base);

	if(type.kind==BasicTypeDeclaration::VECTOR && type.base_type)
		if(BasicTypeDeclaration *basic_base = dynamic_cast<BasicTypeDeclaration *>(type.base_type))
			if(basic_base->kind==BasicTypeDeclaration::VECTOR)
			{
				type.kind = BasicTypeDeclaration::MATRIX;
				type.size |= basic_base->size<<16;
			}

	if(type.kind==BasicTypeDeclaration::ALIAS && type.base_type)
		alias_map[&type] = type.base_type;
	else if(type.kind==BasicTypeDeclaration::ARRAY && type.base_type)
		array_types[type.base_type] = &type;

	stage->types.insert(make_pair(type.name, &type));
}

void TypeResolver::visit(ImageTypeDeclaration &type)
{
	type.base_type = resolve_type(type.base);
	stage->types.insert(make_pair(type.name, &type));
}

void TypeResolver::visit(StructDeclaration &strct)
{
	stage->types.insert(make_pair(strct.name, &strct));
	TraversingVisitor::visit(strct);
}

void TypeResolver::visit(VariableDeclaration &var)
{
	TypeDeclaration *type = resolve_type(var.type);
	if(var.array && type)
	{
		map<TypeDeclaration *, TypeDeclaration *>::iterator i = array_types.find(type);
		if(i==array_types.end())
		{
			BasicTypeDeclaration *array = new BasicTypeDeclaration;
			array->source = BUILTIN_SOURCE;
			array->name = type->name+"[]";
			array->kind = BasicTypeDeclaration::ARRAY;
			array->base = type->name;
			array->base_type = type;
			stage->content.body.insert(type_insert_point, array);
			array->visit(*this);
			type = array;
		}
		else
			type = i->second;
	}
	var.type_declaration = type;
}

void TypeResolver::visit(FunctionDeclaration &func)
{
	func.return_type_declaration = resolve_type(func.return_type);
	TraversingVisitor::visit(func);
}


VariableResolver::VariableResolver():
	stage(0),
	r_members(0),
	record_target(false),
	r_self_referencing(false),
	r_assignment_target(0)
{ }

void VariableResolver::apply(Stage &s)
{
	stage = &s;
	s.interface_blocks.clear();
	s.content.visit(*this);
}

void VariableResolver::enter(Block &block)
{
	block.variables.clear();
}

void VariableResolver::visit(VariableReference &var)
{
	var.declaration = 0;
	r_members = 0;
	/* Look for variable declarations in the block hierarchy first.  Interface
	blocks are always defined in the top level so we can't accidentally skip
	one. */
	for(Block *block=current_block; (!var.declaration && block); block=block->parent)
	{
		map<string, VariableDeclaration *>::iterator i = block->variables.find(var.name);
		if(i!=block->variables.end())
			var.declaration = i->second;
	}

	if(!var.declaration)
	{
		const map<string, InterfaceBlock *> &blocks = stage->interface_blocks;
		map<string, InterfaceBlock *>::const_iterator i = blocks.find("_"+var.name);
		if(i!=blocks.end())
		{
			/* The name refers to an interface block with an instance name rather
			than a variable.  Prepare a new syntax tree node accordingly. */
			r_iface_ref = new InterfaceBlockReference;
			r_iface_ref->source = var.source;
			r_iface_ref->line = var.line;
			r_iface_ref->name = var.name;
			r_iface_ref->declaration = i->second;
			r_members = &i->second->members.variables;
		}
		else
		{
			// Look for the variable in anonymous interface blocks.
			for(i=blocks.begin(); (!var.declaration && i!=blocks.end()); ++i)
				if(i->second->instance_name.empty())
				{
					map<string, VariableDeclaration *>::iterator j = i->second->members.variables.find(var.name);
					if(j!=i->second->members.variables.end())
						var.declaration = j->second;
				}
		}
	}

	if(var.declaration)
		if(StructDeclaration *strct = dynamic_cast<StructDeclaration *>(var.declaration->type_declaration))
			r_members = &strct->members.variables;

	if(record_target)
	{
		if(r_assignment_target)
		{
			/* More than one variable reference found in assignment target.
			Unable to determine what the primary target is. */
			record_target = false;
			r_assignment_target = 0;
		}
		else
			r_assignment_target = var.declaration;
	}
	else if(var.declaration && var.declaration==r_assignment_target)
		r_self_referencing = true;
}

void VariableResolver::visit(InterfaceBlockReference &iface)
{
	iface.declaration = 0;
	for(Block *block=current_block; block; block=block->parent)
	{
		map<string, InterfaceBlock *>::iterator i = stage->interface_blocks.find("_"+iface.name);
		if(i!=stage->interface_blocks.end())
		{
			iface.declaration = i->second;
			r_members = &i->second->members.variables;
			break;
		}
	}
}

void VariableResolver::visit(MemberAccess &memacc)
{
	r_members = 0;
	r_iface_ref = 0;
	memacc.left->visit(*this);

	if(r_iface_ref)
		memacc.left = r_iface_ref;
	r_iface_ref = 0;

	memacc.declaration = 0;
	if(r_members)
	{
		map<string, VariableDeclaration *>::iterator i = r_members->find(memacc.member);
		if(i!=r_members->end())
		{
			memacc.declaration = i->second;
			if(StructDeclaration *strct = dynamic_cast<StructDeclaration *>(i->second->type_declaration))
				r_members = &strct->members.variables;
		}
		else
			r_members = 0;
	}
}

void VariableResolver::visit(UnaryExpression &unary)
{
	TraversingVisitor::visit(unary);
	r_members = 0;
	r_iface_ref = 0;
}

void VariableResolver::visit(BinaryExpression &binary)
{
	if(binary.oper->token[0]=='[')
	{
		{
			/* The subscript expression is not a part of the primary assignment
			target. */
			SetFlag set(record_target, false);
			binary.right->visit(*this);
		}
		r_members = 0;
		r_iface_ref = 0;
		binary.left->visit(*this);
		if(r_iface_ref)
			binary.left = r_iface_ref;
	}
	else
	{
		TraversingVisitor::visit(binary);
		r_members = 0;
	}

	r_iface_ref = 0;
}

void VariableResolver::visit(Assignment &assign)
{
	{
		SetFlag set(record_target);
		r_assignment_target = 0;
		assign.left->visit(*this);
		assign.target_declaration = r_assignment_target;
	}

	r_self_referencing = false;
	assign.right->visit(*this);
	assign.self_referencing = (r_self_referencing || assign.oper->token[0]!='=');

	r_members = 0;
	r_iface_ref = 0;
}

void VariableResolver::visit(FunctionCall &call)
{
	TraversingVisitor::visit(call);
	r_members = 0;
	r_iface_ref = 0;
}

void VariableResolver::visit(VariableDeclaration &var)
{
	if(!block_interface.empty() && var.interface.empty())
		var.interface = block_interface;

	TraversingVisitor::visit(var);
	current_block->variables.insert(make_pair(var.name, &var));
}

void VariableResolver::visit(InterfaceBlock &iface)
{
	/* Block names can be reused in different interfaces.  Prefix the name with
	the first character of the interface to avoid conflicts. */
	stage->interface_blocks.insert(make_pair(iface.interface+iface.name, &iface));
	if(!iface.instance_name.empty())
		stage->interface_blocks.insert(make_pair("_"+iface.instance_name, &iface));

	SetForScope<string> set_iface(block_interface, iface.interface);
	TraversingVisitor::visit(iface);
}


bool ExpressionResolver::is_scalar(BasicTypeDeclaration &type)
{
	return (type.kind==BasicTypeDeclaration::INT || type.kind==BasicTypeDeclaration::FLOAT);
}

bool ExpressionResolver::is_vector_or_matrix(BasicTypeDeclaration &type)
{
	return (type.kind==BasicTypeDeclaration::VECTOR || type.kind==BasicTypeDeclaration::MATRIX);
}

BasicTypeDeclaration *ExpressionResolver::get_element_type(BasicTypeDeclaration &type)
{
	if(is_vector_or_matrix(type) || type.kind==BasicTypeDeclaration::ARRAY)
	{
		BasicTypeDeclaration *basic_base = dynamic_cast<BasicTypeDeclaration *>(type.base_type);
		return (basic_base ? get_element_type(*basic_base) : 0);
	}
	else
		return &type;
}

bool ExpressionResolver::can_convert(BasicTypeDeclaration &from, BasicTypeDeclaration &to)
{
	if(from.kind==BasicTypeDeclaration::INT && to.kind==BasicTypeDeclaration::FLOAT)
		return from.size<=to.size;
	else if(from.kind!=to.kind)
		return false;
	else if((from.kind==BasicTypeDeclaration::VECTOR || from.kind==BasicTypeDeclaration::MATRIX) && from.size==to.size)
	{
		BasicTypeDeclaration *from_base = dynamic_cast<BasicTypeDeclaration *>(from.base_type);
		BasicTypeDeclaration *to_base = dynamic_cast<BasicTypeDeclaration *>(to.base_type);
		return (from_base && to_base && can_convert(*from_base, *to_base));
	}
	else
		return false;
}

ExpressionResolver::Compatibility ExpressionResolver::get_compatibility(BasicTypeDeclaration &left, BasicTypeDeclaration &right)
{
	if(&left==&right)
		return SAME_TYPE;
	else if(can_convert(left, right))
		return LEFT_CONVERTIBLE;
	else if(can_convert(right, left))
		return RIGHT_CONVERTIBLE;
	else
		return NOT_COMPATIBLE;
}

BasicTypeDeclaration *ExpressionResolver::find_type(BasicTypeDeclaration::Kind kind, unsigned size)
{
	for(vector<BasicTypeDeclaration *>::const_iterator i=basic_types.begin(); i!=basic_types.end(); ++i)
		if((*i)->kind==kind && (*i)->size==size)
			return *i;
	return 0;
}

BasicTypeDeclaration *ExpressionResolver::find_type(BasicTypeDeclaration &elem_type, BasicTypeDeclaration::Kind kind, unsigned size)
{
	for(vector<BasicTypeDeclaration *>::const_iterator i=basic_types.begin(); i!=basic_types.end(); ++i)
		if(get_element_type(**i)==&elem_type && (*i)->kind==kind && (*i)->size==size)
			return *i;
	return 0;
}

void ExpressionResolver::convert_to(RefPtr<Expression> &expr, BasicTypeDeclaration &type)
{
	RefPtr<FunctionCall> call = new FunctionCall;
	call->name = type.name;
	call->constructor = true;
	call->arguments.push_back(0);
	call->arguments.back() = expr;
	call->type = &type;
	expr = call;
}

bool ExpressionResolver::convert_to_element(RefPtr<Expression> &expr, BasicTypeDeclaration &elem_type)
{
	if(BasicTypeDeclaration *expr_type = dynamic_cast<BasicTypeDeclaration *>(expr->type))
	{
		BasicTypeDeclaration *to_type = &elem_type;
		if(is_vector_or_matrix(*expr_type))
			to_type = find_type(elem_type, expr_type->kind, expr_type->size);
		if(to_type)
		{
			convert_to(expr, *to_type);
			return true;
		}
	}

	return false;
}

void ExpressionResolver::visit(Literal &literal)
{
	if(literal.value.check_type<bool>())
		literal.type = find_type(BasicTypeDeclaration::BOOL, 1);
	else if(literal.value.check_type<int>())
		literal.type = find_type(BasicTypeDeclaration::INT, 32);
	else if(literal.value.check_type<float>())
		literal.type = find_type(BasicTypeDeclaration::FLOAT, 32);
}

void ExpressionResolver::visit(ParenthesizedExpression &parexpr)
{
	TraversingVisitor::visit(parexpr);

	parexpr.type = parexpr.expression->type;
	parexpr.lvalue = parexpr.expression->lvalue;
}

void ExpressionResolver::visit(VariableReference &var)
{
	if(var.declaration)
		var.type = var.declaration->type_declaration;
	var.lvalue = true;
}

void ExpressionResolver::visit(InterfaceBlockReference &iface)
{
	iface.lvalue = true;
}

void ExpressionResolver::visit(MemberAccess &memacc)
{
	TraversingVisitor::visit(memacc);

	if(memacc.declaration)
		memacc.type = memacc.declaration->type_declaration;
	memacc.lvalue = memacc.left->lvalue;
}

void ExpressionResolver::visit(UnaryExpression &unary)
{
	TraversingVisitor::visit(unary);

	BasicTypeDeclaration *basic = dynamic_cast<BasicTypeDeclaration *>(unary.expression->type);
	if(!basic)
		return;

	char oper = unary.oper->token[0];
	if(oper=='!')
	{
		if(basic->kind==BasicTypeDeclaration::BOOL)
			unary.type = basic;
	}
	else if(oper=='~')
	{
		if(basic->kind==BasicTypeDeclaration::INT)
			unary.type = basic;
	}
	else if(oper=='+' || oper=='-')
	{
		BasicTypeDeclaration *elem = get_element_type(*basic);
		if(elem && is_scalar(*elem))
			unary.type = basic;
	}
	unary.lvalue = unary.expression->lvalue;
}

void ExpressionResolver::visit(BinaryExpression &binary)
{
	TraversingVisitor::visit(binary);

	/* Binary operators are only defined for basic types (not for image or
	structure types). */
	BasicTypeDeclaration *basic_left = dynamic_cast<BasicTypeDeclaration *>(binary.left->type);
	BasicTypeDeclaration *basic_right = dynamic_cast<BasicTypeDeclaration *>(binary.right->type);
	if(!basic_left || !basic_right)
		return;

	binary.lvalue = false;

	char oper = binary.oper->token[0];
	if(oper=='[')
	{
		/* Subscripting operates on vectors, matrices and arrays, and the right
		operand must be an integer. */
		if((!is_vector_or_matrix(*basic_left) && basic_left->kind!=BasicTypeDeclaration::ARRAY) || basic_right->kind!=BasicTypeDeclaration::INT)
			return;

		binary.type = basic_left->base_type;
		binary.lvalue = binary.left->lvalue;
		return;
	}
	else if(basic_left->kind==BasicTypeDeclaration::ARRAY || basic_right->kind==BasicTypeDeclaration::ARRAY)
		// No other binary operator can be used with arrays.
		return;

	BasicTypeDeclaration *elem_left = get_element_type(*basic_left);
	BasicTypeDeclaration *elem_right = get_element_type(*basic_right);
	if(!elem_left || !elem_right)
		return;

	Compatibility compat = get_compatibility(*basic_left, *basic_right);
	Compatibility elem_compat = get_compatibility(*elem_left, *elem_right);
	if(elem_compat==NOT_COMPATIBLE)
		return;

	char oper2 = binary.oper->token[1];
	if((oper=='<' && oper2!='<') || (oper=='>' && oper2!='>'))
	{
		/* Relational operators compare two scalar integer or floating-point
		values. */
		if(!is_scalar(*elem_left) || !is_scalar(*elem_right) || compat==NOT_COMPATIBLE)
			return;

		binary.type = find_type(BasicTypeDeclaration::BOOL, 1);
	}
	else if((oper=='=' || oper=='!') && oper2=='=')
	{
		// Equality comparison can be done on any compatible types.
		if(compat==NOT_COMPATIBLE)
			return;

		binary.type = find_type(BasicTypeDeclaration::BOOL, 1);
	}
	else if(oper2=='&' || oper2=='|' || oper2=='^')
	{
		// Logical operators can only be applied to booleans.
		if(basic_left->kind!=BasicTypeDeclaration::BOOL || basic_right->kind!=BasicTypeDeclaration::BOOL)
			return;

		binary.type = basic_left;
	}
	else if((oper=='&' || oper=='|' || oper=='^' || oper=='%') && !oper2)
	{
		// Bitwise operators and modulo can only be applied to integers.
		if(basic_left->kind!=BasicTypeDeclaration::INT || basic_right->kind!=BasicTypeDeclaration::INT)
			return;

		binary.type = (compat==LEFT_CONVERTIBLE ? basic_right : basic_left);
	}
	else if((oper=='<' || oper=='>') && oper2==oper)
	{
		// Shifts only apply to integers.
		if(basic_left->kind!=BasicTypeDeclaration::INT || basic_right->kind!=BasicTypeDeclaration::INT)
			return;

		binary.type = basic_left;
	}
	else if(oper=='+' || oper=='-' || oper=='*' || oper=='/')
	{
		// Arithmetic operators require scalar elements.
		if(!is_scalar(*elem_left) || !is_scalar(*elem_right))
			return;

		if(oper=='*' && is_vector_or_matrix(*basic_left) && is_vector_or_matrix(*basic_right) &&
			(basic_left->kind==BasicTypeDeclaration::MATRIX || basic_right->kind==BasicTypeDeclaration::MATRIX))
		{
			/* Multiplication has special rules when at least one operand is a
			matrix and the other is a vector or a matrix. */
			unsigned left_columns = basic_left->size&0xFFFF;
			unsigned right_rows = basic_right->size;
			if(basic_right->kind==BasicTypeDeclaration::MATRIX)
				right_rows >>= 16;
			if(left_columns!=right_rows)
				return;

			BasicTypeDeclaration *elem_result = (elem_compat==LEFT_CONVERTIBLE ? elem_right : elem_left);

			if(basic_left->kind==BasicTypeDeclaration::VECTOR)
				binary.type = find_type(*elem_result, BasicTypeDeclaration::VECTOR, basic_right->size&0xFFFF);
			else if(basic_right->kind==BasicTypeDeclaration::VECTOR)
				binary.type = find_type(*elem_result, BasicTypeDeclaration::VECTOR, basic_left->size>>16);
			else
				binary.type = find_type(*elem_result, BasicTypeDeclaration::MATRIX, (basic_left->size&0xFFFF0000)|(basic_right->size&0xFFFF));
		}
		else if(compat==NOT_COMPATIBLE)
		{
			// Arithmetic between scalars and matrices or vectors is supported.
			if(is_scalar(*basic_left) && is_vector_or_matrix(*basic_right))
				binary.type = (elem_compat==RIGHT_CONVERTIBLE ? find_type(*elem_left, basic_right->kind, basic_right->size) : basic_right);
			else if(is_vector_or_matrix(*basic_left) && is_scalar(*basic_right))
				binary.type = (elem_compat==LEFT_CONVERTIBLE ? find_type(*elem_right, basic_left->kind, basic_left->size) : basic_left);
			else
				return;
		}
		else if(compat==LEFT_CONVERTIBLE)
			binary.type = basic_right;
		else
			binary.type = basic_left;
	}
	else
		return;

	bool converted = true;
	if(compat==LEFT_CONVERTIBLE)
		convert_to(binary.left, *basic_right);
	else if(compat==RIGHT_CONVERTIBLE)
		convert_to(binary.right, *basic_left);
	else if(elem_compat==LEFT_CONVERTIBLE)
		converted = convert_to_element(binary.left, *elem_right);
	else if(elem_compat==RIGHT_CONVERTIBLE)
		converted = convert_to_element(binary.right, *elem_left);

	if(!converted)
		binary.type = 0;
}

void ExpressionResolver::visit(Assignment &assign)
{
	TraversingVisitor::visit(assign);
	assign.type = assign.left->type;
	assign.lvalue = true;
}

void ExpressionResolver::visit(FunctionCall &call)
{
	TraversingVisitor::visit(call);

	if(call.declaration)
		call.type = call.declaration->return_type_declaration;
	else if(call.constructor)
	{
		map<string, TypeDeclaration *>::const_iterator i=stage->types.find(call.name);
		call.type = (i!=stage->types.end() ? i->second : 0);
	}
	call.lvalue = false;
}

void ExpressionResolver::visit(BasicTypeDeclaration &type)
{
	basic_types.push_back(&type);
}

void ExpressionResolver::visit(VariableDeclaration &var)
{
	TraversingVisitor::visit(var);
	if(!var.init_expression)
		return;

	BasicTypeDeclaration *var_basic = dynamic_cast<BasicTypeDeclaration *>(var.type_declaration);
	BasicTypeDeclaration *init_basic = dynamic_cast<BasicTypeDeclaration *>(var.init_expression->type);
	if(!var_basic || !init_basic)
		return;

	Compatibility compat = get_compatibility(*var_basic, *init_basic);
	if(compat==RIGHT_CONVERTIBLE)
		convert_to(var.init_expression, *var_basic);
}


void FunctionResolver::apply(Stage &s)
{
	stage = &s;
	s.functions.clear();
	s.content.visit(*this);
}

void FunctionResolver::visit(FunctionCall &call)
{
	map<string, FunctionDeclaration *>::iterator i = stage->functions.find(call.name);
	if(i!=stage->functions.end())
		call.declaration = i->second;

	TraversingVisitor::visit(call);
}

void FunctionResolver::visit(FunctionDeclaration &func)
{
	FunctionDeclaration *&stage_decl = stage->functions[func.name];
	vector<FunctionDeclaration *> &decls = declarations[func.name];
	if(func.definition==&func)
	{
		stage_decl = &func;

		// Set all previous declarations to use this definition.
		for(vector<FunctionDeclaration *>::iterator i=decls.begin(); i!=decls.end(); ++i)
		{
			(*i)->definition = func.definition;
			(*i)->body.body.clear();
		}
	}
	else
	{
		func.definition = 0;
		if(!stage_decl)
			stage_decl = &func;
		else
			func.definition = stage_decl->definition;
	}
	decls.push_back(&func);

	TraversingVisitor::visit(func);
}


InterfaceGenerator::InterfaceGenerator():
	stage(0),
	function_scope(false),
	iface_block(0),
	copy_block(false),
	iface_target_block(0)
{ }

string InterfaceGenerator::get_out_prefix(Stage::Type type)
{
	if(type==Stage::VERTEX)
		return "_vs_out_";
	else if(type==Stage::GEOMETRY)
		return "_gs_out_";
	else
		return string();
}

void InterfaceGenerator::apply(Stage &s)
{
	stage = &s;
	iface_target_block = &stage->content;
	if(stage->previous)
		in_prefix = get_out_prefix(stage->previous->type);
	out_prefix = get_out_prefix(stage->type);
	s.content.visit(*this);
	NodeRemover().apply(s, nodes_to_remove);
}

void InterfaceGenerator::visit(Block &block)
{
	SetForScope<Block *> set_block(current_block, &block);
	for(NodeList<Statement>::iterator i=block.body.begin(); i!=block.body.end(); ++i)
	{
		assignment_insert_point = i;
		if(&block==&stage->content)
			iface_insert_point = i;

		(*i)->visit(*this);
	}
}

string InterfaceGenerator::change_prefix(const string &name, const string &prefix) const
{
	unsigned offset = (name.compare(0, in_prefix.size(), in_prefix) ? 0 : in_prefix.size());
	return prefix+name.substr(offset);
}

VariableDeclaration *InterfaceGenerator::generate_interface(VariableDeclaration &var, const string &iface, const string &name)
{
	if(stage->content.variables.count(name))
		return 0;

	VariableDeclaration* iface_var = new VariableDeclaration;
	iface_var->sampling = var.sampling;
	iface_var->interface = iface;
	iface_var->type = var.type;
	iface_var->name = name;
	/* Geometry shader inputs are always arrays.  But if we're bringing in an
	entire block, the array is on the block and not individual variables. */
	if(stage->type==Stage::GEOMETRY && !copy_block)
		iface_var->array = ((var.array && var.interface!="in") || iface=="in");
	else
		iface_var->array = var.array;
	if(iface_var->array)
		iface_var->array_size = var.array_size;
	if(iface=="in")
	{
		iface_var->layout = var.layout;
		iface_var->linked_declaration = &var;
		var.linked_declaration = iface_var;
	}

	iface_target_block->body.insert(iface_insert_point, iface_var);
	iface_target_block->variables.insert(make_pair(name, iface_var));

	return iface_var;
}

InterfaceBlock *InterfaceGenerator::generate_interface(InterfaceBlock &out_block)
{
	if(stage->interface_blocks.count("in"+out_block.name))
		return 0;

	InterfaceBlock *in_block = new InterfaceBlock;
	in_block->interface = "in";
	in_block->name = out_block.name;
	in_block->instance_name = out_block.instance_name;
	if(stage->type==Stage::GEOMETRY)
		in_block->array = true;
	else
		in_block->array = out_block.array;
	in_block->linked_block = &out_block;
	out_block.linked_block = in_block;

	{
		SetFlag set_copy(copy_block, true);
		SetForScope<Block *> set_target(iface_target_block, &in_block->members);
		SetForScope<NodeList<Statement>::iterator> set_ins_pt(iface_insert_point, in_block->members.body.end());
		out_block.members.visit(*this);
	}

	iface_target_block->body.insert(iface_insert_point, in_block);
	stage->interface_blocks.insert(make_pair("in"+in_block->name, in_block));
	if(!in_block->instance_name.empty())
		stage->interface_blocks.insert(make_pair("_"+in_block->instance_name, in_block));

	SetFlag set_scope(function_scope, false);
	SetForScope<Block *> set_block(current_block, &stage->content);
	in_block->visit(*this);

	return in_block;
}

ExpressionStatement &InterfaceGenerator::insert_assignment(const string &left, Expression *right)
{
	Assignment *assign = new Assignment;
	VariableReference *ref = new VariableReference;
	ref->name = left;
	assign->left = ref;
	assign->oper = &Operator::get_operator("=", Operator::BINARY);
	assign->right = right;

	ExpressionStatement *stmt = new ExpressionStatement;
	stmt->expression = assign;
	current_block->body.insert(assignment_insert_point, stmt);
	stmt->visit(*this);

	return *stmt;
}

void InterfaceGenerator::visit(VariableReference &var)
{
	if(var.declaration || !stage->previous)
		return;
	/* Don't pull a variable from previous stage if we just generated an output
	interface in this stage */
	if(stage->content.variables.count(var.name))
		return;

	const map<string, VariableDeclaration *> &prev_vars = stage->previous->content.variables;
	map<string, VariableDeclaration *>::const_iterator i = prev_vars.find(var.name);
	if(i==prev_vars.end() || i->second->interface!="out")
		i = prev_vars.find(in_prefix+var.name);
	if(i!=prev_vars.end() && i->second->interface=="out")
	{
		generate_interface(*i->second, "in", i->second->name);
		var.name = i->second->name;
		return;
	}

	const map<string, InterfaceBlock *> &prev_blocks = stage->previous->interface_blocks;
	map<string, InterfaceBlock *>::const_iterator j = prev_blocks.find("_"+var.name);
	if(j!=prev_blocks.end() && j->second->interface=="out")
	{
		generate_interface(*j->second);
		/* Let VariableResolver convert the variable reference into an interface
		block reference. */
		return;
	}

	for(j=prev_blocks.begin(); j!=prev_blocks.end(); ++j)
		if(j->second->instance_name.empty())
		{
			i = j->second->members.variables.find(var.name);
			if(i!=j->second->members.variables.end())
			{
				generate_interface(*j->second);
				return;
			}
		}
}

void InterfaceGenerator::visit(VariableDeclaration &var)
{
	if(copy_block)
	{
		generate_interface(var, "in", var.name);
		return;
	}

	if(iface_block)
	{
		if(iface_block->linked_block)
		{
			// Link all variables to their counterparts in the linked block.
			const map<string, VariableDeclaration *> &linked_vars = iface_block->linked_block->members.variables;
			map<string, VariableDeclaration *>::const_iterator i = linked_vars.find(var.name);
			if(i!=linked_vars.end())
			{
				var.linked_declaration = i->second;
				var.linked_declaration->linked_declaration = &var;
			}
		}
		return;
	}

	if(var.interface=="out")
	{
		/* For output variables in function scope, generate a global interface
		and replace the local declaration with an assignment. */
		VariableDeclaration *out_var = 0;
		if(function_scope && (out_var=generate_interface(var, "out", var.name)))
		{
			out_var->source = var.source;
			out_var->line = var.line;
			nodes_to_remove.insert(&var);
			if(var.init_expression)
			{
				ExpressionStatement &stmt = insert_assignment(var.name, var.init_expression->clone());
				stmt.source = var.source;
				stmt.line = var.line;
				return;
			}
		}
	}
	else if(var.interface=="in")
	{
		/* Try to link input variables in global scope with output variables from
		previous stage. */
		if(current_block==&stage->content && !var.linked_declaration && stage->previous)
		{
			const map<string, VariableDeclaration *> &prev_vars = stage->previous->content.variables;
			map<string, VariableDeclaration *>::const_iterator i = prev_vars.find(var.name);
			if(i!=prev_vars.end() && i->second->interface=="out")
			{
				var.linked_declaration = i->second;
				i->second->linked_declaration = &var;
			}
		}
	}

	TraversingVisitor::visit(var);
}

void InterfaceGenerator::visit(InterfaceBlock &iface)
{
	if(iface.interface=="in")
	{
		/* Try to link input blocks with output blocks sharing the same block
		name from previous stage. */
		if(!iface.linked_block && stage->previous)
		{
			const map<string, InterfaceBlock *> &prev_blocks = stage->previous->interface_blocks;
			map<string, InterfaceBlock *>::const_iterator i = prev_blocks.find("out"+iface.name);
			if(i!=prev_blocks.end())
			{
				iface.linked_block = i->second;
				i->second->linked_block = &iface;
			}
		}
	}

	SetForScope<InterfaceBlock *> set_iface(iface_block, &iface);
	TraversingVisitor::visit(iface);
}

void InterfaceGenerator::visit(FunctionDeclaration &func)
{
	SetFlag set_scope(function_scope, true);
	// Skip parameters because they're not useful here
	func.body.visit(*this);
}

void InterfaceGenerator::visit(Passthrough &pass)
{
	vector<VariableDeclaration *> pass_vars;

	// Pass through all input variables of this stage.
	for(map<string, VariableDeclaration *>::const_iterator i=stage->content.variables.begin(); i!=stage->content.variables.end(); ++i)
		if(i->second->interface=="in")
			pass_vars.push_back(i->second);

	if(stage->previous)
	{
		const map<string, VariableDeclaration *> &prev_vars = stage->previous->content.variables;
		for(map<string, VariableDeclaration *>::const_iterator i=prev_vars.begin(); i!=prev_vars.end(); ++i)
		{
			if(i->second->interface!="out")
				continue;

			/* Pass through output variables from the previous stage, but only
			those which are not already linked to an input here. */
			if(!i->second->linked_declaration && generate_interface(*i->second, "in", i->second->name))
				pass_vars.push_back(i->second);
		}
	}

	if(stage->type==Stage::GEOMETRY)
	{
		/* Special case for geometry shader: copy gl_Position from input to
		output. */
		InterfaceBlockReference *ref = new InterfaceBlockReference;
		ref->name = "gl_in";

		BinaryExpression *subscript = new BinaryExpression;
		subscript->left = ref;
		subscript->oper = &Operator::get_operator("[", Operator::BINARY);
		subscript->right = pass.subscript;

		MemberAccess *memacc = new MemberAccess;
		memacc->left = subscript;
		memacc->member = "gl_Position";

		insert_assignment("gl_Position", memacc);
	}

	for(vector<VariableDeclaration *>::const_iterator i=pass_vars.begin(); i!=pass_vars.end(); ++i)
	{
		string out_name = change_prefix((*i)->name, out_prefix);
		generate_interface(**i, "out", out_name);

		VariableReference *ref = new VariableReference;
		ref->name = (*i)->name;
		if(pass.subscript)
		{
			BinaryExpression *subscript = new BinaryExpression;
			subscript->left = ref;
			subscript->oper = &Operator::get_operator("[", Operator::BINARY);
			subscript->right = pass.subscript;
			insert_assignment(out_name, subscript);
		}
		else
			insert_assignment(out_name, ref);
	}

	nodes_to_remove.insert(&pass);
}

} // namespace SL
} // namespace GL
} // namespace Msp
