#include <msp/core/raii.h>
#include "visitor.h"

using namespace std;

namespace Msp {
namespace GL {
namespace SL {

void TraversingVisitor::visit(Block &block)
{
	if(&block!=current_block)
		enter(block);
	SetForScope<Block *> set_block(current_block, &block);
	for(NodeList<Statement>::iterator i=block.body.begin(); i!=block.body.end(); ++i)
		(*i)->visit(*this);
}

void TraversingVisitor::visit(ParenthesizedExpression &parexpr)
{
	parexpr.expression->visit(*this);
}

void TraversingVisitor::visit(MemberAccess &memacc)
{
	memacc.left->visit(*this);
}

void TraversingVisitor::visit(UnaryExpression &unary)
{
	unary.expression->visit(*this);
}

void TraversingVisitor::visit(BinaryExpression &binary)
{
	binary.left->visit(*this);
	binary.right->visit(*this);
}

void TraversingVisitor::visit(Assignment &assign)
{
	visit(static_cast<BinaryExpression &>(assign));
}

void TraversingVisitor::visit(FunctionCall &call)
{
	for(NodeArray<Expression>::iterator i=call.arguments.begin(); i!=call.arguments.end(); ++i)
		(*i)->visit(*this);
}

void TraversingVisitor::visit(ExpressionStatement &expr)
{
	expr.expression->visit(*this);
}

void TraversingVisitor::visit(InterfaceLayout &layout)
{
	layout.layout.visit(*this);
}

void TraversingVisitor::visit(StructDeclaration &strct)
{
	strct.members.visit(*this);
}

void TraversingVisitor::visit(VariableDeclaration &var)
{
	if(var.layout)
		var.layout->visit(*this);
	if(var.init_expression)
		var.init_expression->visit(*this);
	if(var.array_size)
		var.array_size->visit(*this);
}

void TraversingVisitor::visit(InterfaceBlock &iface)
{
	iface.members.visit(*this);
}

void TraversingVisitor::visit(FunctionDeclaration &func)
{
	enter(func.body);
	SetForScope<Block *> set_block(current_block, &func.body);
	for(NodeArray<VariableDeclaration>::iterator i=func.parameters.begin(); i!=func.parameters.end(); ++i)
		(*i)->visit(*this);
	func.body.visit(*this);
}

void TraversingVisitor::visit(Conditional &cond)
{
	cond.condition->visit(*this);
	cond.body.visit(*this);
	cond.else_body.visit(*this);
}

void TraversingVisitor::visit(Iteration &iter)
{
	enter(iter.body);
	SetForScope<Block *> set_block(current_block, &iter.body);
	if(iter.init_statement)
		iter.init_statement->visit(*this);
	if(iter.condition)
		iter.condition->visit(*this);
	iter.body.visit(*this);
	if(iter.loop_expression)
		iter.loop_expression->visit(*this);
}

void TraversingVisitor::visit(Passthrough &pass)
{
	if(pass.subscript)
		pass.subscript->visit(*this);
}

void TraversingVisitor::visit(Return &ret)
{
	if(ret.expression)
		ret.expression->visit(*this);
}


NodeRemover::NodeRemover():
	stage(0),
	to_remove(0),
	recursive_remove(false)
{ }

void NodeRemover::apply(Stage &s, const set<Node *> &tr)
{
	stage = &s;
	to_remove = &tr;
	s.content.visit(*this);
}

template<typename T>
void NodeRemover::remove_from_map(map<string, T *> &vars, const string &key, T &node)
{
	typename map<string, T *>::iterator i = vars.find(key);
	if(i!=vars.end() && i->second==&node)
		vars.erase(i);
}

void NodeRemover::visit(Block &block)
{
	SetForScope<Block *> set_block(current_block, &block);
	for(NodeList<Statement>::iterator i=block.body.begin(); i!=block.body.end(); )
	{
		(*i)->visit(*this);
		if(to_remove->count(i->get()))
			block.body.erase(i++);
		else
			++i;
	}
}

void NodeRemover::visit(TypeDeclaration &type)
{
	if(to_remove->count(&type))
		remove_from_map(stage->types, type.name, type);
}

void NodeRemover::visit(VariableDeclaration &var)
{
	if(recursive_remove || to_remove->count(&var))
	{
		remove_from_map(current_block->variables, var.name, var);
		stage->locations.erase(var.name);
		if(var.linked_declaration)
			var.linked_declaration->linked_declaration = 0;
	}
	else if(var.init_expression && to_remove->count(var.init_expression.get()))
		var.init_expression = 0;
}

void NodeRemover::visit(InterfaceBlock &iface)
{
	if(to_remove->count(&iface))
	{
		remove_from_map(stage->interface_blocks, iface.interface+iface.name, iface);
		if(!iface.instance_name.empty())
			remove_from_map(stage->interface_blocks, "_"+iface.instance_name, iface);
	}
	SetFlag set_recursive(recursive_remove, recursive_remove || to_remove->count(&iface));
	TraversingVisitor::visit(iface);
}

void NodeRemover::visit(FunctionDeclaration &func)
{
	if(to_remove->count(&func))
		remove_from_map(stage->functions, func.name, func);
	TraversingVisitor::visit(func);
}

void NodeRemover::visit(Iteration &iter)
{
	if(to_remove->count(iter.init_statement.get()))
		iter.init_statement = 0;
	TraversingVisitor::visit(iter);
}


NodeReorderer::NodeReorderer():
	reorder_before(0),
	to_reorder(0)
{ }

void NodeReorderer::apply(Stage &stage, Node &before, const set<Node *> &tr)
{
	reorder_before = &before;
	to_reorder = &tr;
	stage.content.visit(*this);
}

void NodeReorderer::visit(Block &block)
{
	NodeList<Statement>::iterator insert_point = block.body.end();
	for(NodeList<Statement>::iterator i=block.body.begin(); i!=block.body.end(); )
	{
		(*i)->visit(*this);
		if(insert_point!=block.body.end() && to_reorder->count(i->get()))
		{
			NodeList<Statement>::iterator j = i++;
			block.body.splice(insert_point, block.body, j);
		}
		else
		{
			if(i->get()==reorder_before)
				insert_point = i;
			++i;
		}
	}
}

} // namespace SL
} // namespace GL
} // namespace Msp
